/*
 * sample program for using Qrcode class.
 * (c) Copyright 2005 by NAKATA H.<webmaster@nakata-jp.org>.
 */

#include "qrcode.h"

#define	ENV_DATA_PATH	"QRCODE_DATA_PATH" // environment show data directory
#define	DEFAULT_MAGNITUDE	8
#define	DEFAULT_ECCLEVEL	'M'

static void	usage( const char *);
static int	iskanji( const char * );
static void	bmpout( const char *, const char *, int, int );
static void	longout( std::ofstream &, long);
static void	shortout( std::ofstream &, short);
static void	bitout( std::ofstream &, int );
static void	bitflush( std::ofstream & );

static char	code_type[256] = {
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  6, 4, 4, 4, 6, 6, 4, 4, 4, 4, 6, 6, 4, 6, 6, 6,
  7, 7, 7, 7, 7, 7, 7, 7, 7, 7, 6, 4, 4, 4, 4, 4,
  6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 6,
  6, 6, 6, 6, 6, 6, 6, 6, 6, 6, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4};

int	main( int argc, char *argv[] )
{
  int	argcount;
  char	eclevel = DEFAULT_ECCLEVEL;
  int	magnitude = DEFAULT_MAGNITUDE;

  for ( argcount = 1;
	argcount < argc && argv[argcount][0] == '-'; argcount++ ) {
    switch ( argv[argcount][1] ) {
    case 'm':
      if ( argcount >= argc - 1 ) {
	break;
      }
      magnitude = atoi(argv[++argcount]);
      break;
    case 'H':
    case 'M':
    case 'Q':
    case 'L':
      eclevel = argv[argcount][1];
      break;
    default:
      usage( argv[0] );
      break;			// not reached
    }
  }
    
  if ( argc - argcount != 2 ) {
    usage(argv[0]);
  }

  char	*data = argv[argcount];
  char	*outfile = argv[argcount + 1];

  Qrcode qrcode;
  char	*env;

  qrcode.setQrcodeErrorCorrect(eclevel);

  if ( (env = getenv(ENV_DATA_PATH)) != NULL ) {
    if ( !qrcode.setQrcodeDataPath(env) ) {
      std::cerr << "illegal data path " << env << "\n";
    }
  }

  int	code = 0xff;
  for ( unsigned int i = 0; i < strlen(data); i++ ) {
    code &= code_type[data[i] & 0xff];
  }
  if ( (code & 1) == 1 ) {
    qrcode.setQrcodeEncodeMode('N');
  } else if ( (code & 2) == 2 ) {
    qrcode.setQrcodeEncodeMode('A');
  } else if ( iskanji( data ) ) {
    qrcode.setQrcodeEncodeMode('K');
  } else {
    qrcode.setQrcodeEncodeMode('8');
  }

  char	*result;
  
  if ( (result = qrcode.calQrcode( data, strlen(data) )) == NULL ) {
    std::cerr << "convert failed.\n";
    exit(1);
  }

  bmpout( outfile, result, qrcode.modules1Side, magnitude );
#ifdef	DEBUG
  for ( int row = 0; row < qrcode.modules1Side; row++ ) {
    for ( int col = 0; col < qrcode.modules1Side; col++ ) {
      switch (result[col + row * qrcode.modules1Side] ) {
      case 0:
	putchar('*');
	break;
      default:
	putchar('.');
	break;
      }
    }
    putchar('\n');
  }
#endif
  return (0);
}

static int	iskanji( const char *string )
{
  const unsigned char	*p;

  for ( p = (const unsigned char *)string; *p != '\0'; ) {
    if ( *p < 0x81 || 0xeb < *p ) {
      return (FALSE);
    } else if ( 0x9f < *p && *p < 0xe0 ) {
      return (FALSE);
    }
    p++;
    if ( *p < 0x40 || 0xfc < *p ) {
      return (FALSE);
    }
    p++;
  }
  return (TRUE);
}

static void	bmpout( const char *filename, const char *data,
			int size, int magnitude )
{
  long	width, height, filesize;

  width = height = (size + 8) * magnitude;
  std::ofstream file;
  file.open(filename, std::ios_base::out | std::ios_base::binary);

  filesize = (width + 31) / 32 * 4;
  filesize = filesize * height + 62;
  file.write( "BM", 2 );
  longout( file, filesize );
  shortout( file, 0 );
  shortout( file, 0 );
  longout( file, 62 );		// offset of data
  longout( file, 0x28 );
  longout( file, width );
  longout( file, height );
  shortout( file, 1 );	// 1 plane / file
  shortout( file, 1 );	// 1 bit / pixel
  longout( file, 0 );	// no compression
  longout( file, 0 );	// need not to save in this case
  longout( file, 0 );	// need not to save in this case
  longout( file, 0 );	// need not to save in this case
  longout( file, 2 );	// use 2 colors
  longout( file, 0 );	// all colors are important
  longout( file, 0xffffff );	// color palette white
  longout( file, 0 );		// color palette black

  int	x, y;
  for ( int row = height - 1; row >= 0; row-- ) {
    y = row / magnitude - 4;
    for ( int col = 0; col < width; col++ ) {
      x = col / magnitude - 4;
      if ( x < 0 || y < 0 || x >= size || y >= size ) {
	bitout( file, 0 );
	continue;
      }
      switch ( data[ x + y * size ] ) {
      case 0:
	bitout( file, 1 );
	break;
      default:
	bitout( file, 0 );
	break;
      }
    }
    bitflush( file );
  }

  file.close();
}

static void	longout( std::ofstream &file, long l)
{
  char	c;
  c = l & 0xff;
  file.write( &c, 1 );
  c = (l >> 8) & 0xff;
  file.write( &c, 1 );
  c = (l >> 16) & 0xff;
  file.write( &c, 1 );
  c = (l >> 24) & 0xff;
  file.write( &c, 1 );
}

static void	shortout( std::ofstream &file, short s)
{
  char	c;
  c = s & 0xff;
  file.write( &c, 1 );
  c = (s >> 8) & 0xff;
  file.write( &c, 1 );
}

static long	bits = 0;
static int	bitcount = 0;

static void	bitout( std::ofstream &file, int b )
{

  bits = (bits << 1) | b;
  if ( ++bitcount < 32 ) {
    return;
  }
  char	c;
  c = (bits >> 24) & 0xff;
  file.write( &c, 1 );
  c = (bits >> 16) & 0xff;
  file.write( &c, 1 );
  c = (bits >> 8) & 0xff;
  file.write( &c, 1 );
  c = bits & 0xff;
  file.write( &c, 1 );
 
  bits = 0;
  bitcount = 0;
}

static void	bitflush( std::ofstream &file )
{
  while ( bitcount > 0 ) {
    bitout( file, 0 );
  }
}

static void	usage( const char *command )
{
  std::cerr << "qrcode: QR code encoder.\n";
  std::cerr << "usage: " << command << " [-m num] [-H] [-M] [-Q] [-L] data outfile\n";
  std::cerr << "\t-m num: output size 1block = num x num pixels\n";
  std::cerr << "\t-H: set Error Correction Level H\n";
  std::cerr << "\t-M: set Error Correction Level M\n";
  std::cerr << "\t-Q: set Error Correction Level Q\n";
  std::cerr << "\t-L: set Error Correction Level Lp\n";
  std::cerr << "\tdata is string to encode.\n";
  std::cerr << "\toutfile is file name to output(BMP format)\n";
  exit(1);
}
/*	End Of File	******************************************************/
